from typing import Any
from django.http import HttpRequest
from django.http.response import HttpResponse as HttpResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.urls import reverse, reverse_lazy
from django.views import generic

from webapp.forms import ArticleForm, CommentForm
from webapp.models import Article


class ArticleDetailView(generic.DetailView):
    pk_url_kwarg = 'id'
    model = Article
    template_name = 'articles/detail.html'
    
    def get_context_data(self, **kwargs):
        comments = self.object.comments.order_by('-created_at')
        return super().get_context_data(
            **kwargs, 
            comments=comments,
            comments_form=CommentForm,
        )


class ArticleCreateView(generic.CreateView):
    model = Article
    template_name = 'articles/article.html'
    form_class = ArticleForm

    def get_success_url(self):
        return reverse('articles_detail', kwargs={'id': self.object.id})


class UpdateView(generic.View):
    template_name = None
    model = None
    form_class = None
    context_object_name = None
    pk_url_kwarg = None
    success_url = None
    
    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        form = self.form_class(instance=self.object)
        return self.render_to_response(form=form)
    
    def post(self, request, *args, **kwargs):
        form = self.form_class(instance=self.object, data=request.POST)
        
        if form.is_valid():
            return self.form_valid(form)

        return self.form_invalid(form)

    def get_object(self):
        return get_object_or_404(self.model, id=self.kwargs.get(self.pk_url_kwarg))
    
    def get_context_data(self, **kwargs):
        kwargs[self.context_object_name] = self.object
        return kwargs

    def get_success_url(self):
        return self.success_url

    def form_valid(self, form):
        self.object = form.save()
        return redirect(self.get_success_url())
    
    def form_invalid(self, form):
        return self.render_to_response(form=form)

    def render_to_response(self, **kwargs):
        return render(
            self.request, 
            self.template_name, 
            context=self.get_context_data(**kwargs),
        )


class ArticleUpdateView(generic.UpdateView):
    template_name = 'articles/article.html'
    model = Article
    form_class = ArticleForm
    context_object_name = 'article'
    pk_url_kwarg = 'id'
    extra_context = {'btn_text': 'Update'}

    def get_success_url(self):
        return reverse('articles_detail', kwargs={'id': self.object.id})


class DeleteView(generic.View):
    template_name = None
    model = None
    context_object_name = None
    pk_url_kwarg = None
    success_url = None
    
    def dispatch(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        self.object = self.get_object()
        return super().dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        return self.render_to_response()

    def post(self, request, *args, **kwargs):
        self.perform_delete()
        return redirect(self.get_success_url())

    def perform_delete(self):
        self.object.delete()

    def get_object(self):
        return get_object_or_404(self.model, id=self.kwargs.get(self.pk_url_kwarg))
    
    def get_context_data(self, **kwargs):
        kwargs[self.context_object_name] = self.object
        return kwargs

    def get_success_url(self):
        return self.success_url

    def render_to_response(self, **kwargs):
        return render(
            self.request, 
            self.template_name, 
            context=self.get_context_data(**kwargs),
        )


class ArticleDeleteView(generic.DeleteView):
    model = Article
    pk_url_kwarg = 'id'
    success_url = reverse_lazy('index')
